/*
 * dcs_io.c
 *
 * Copyright 2019-2023 Ichiji Tadokoro. All Rights Reserved.
 */

#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include "_dcs.h"
#include "dcs_io.h"

#define _SKIP_BOM		0

#define _DST_NOTHING	0
#define _DST_DIR		1
#define _DST_FILE		2
#define _DST_OTHER		9

static long _is_decided;
static long _stype;
static long _ope_type;

typedef struct _text_io_info {
	Dcs *dcs;
	char *buf;
	size_t bsize; // データ・サイズ
	size_t _blength; // 実際のサイズ
	char *bp;
	long mode;
	long stype;
	long dtype;
	int shtype;
} TioInfo;

static TioInfo _tioInfo[DCS_MAX_TIO];

#define _check_fd(fd)	\
		if (fd < 0 || fd >= DCS_MAX_TIO) {	\
			return DCS_EIOFDRANGE;	\
		} else if (_tioInfo[fd].buf == NULL) {	\
			return DCS_EIONOTOPEN;	\
		}

/* Dcs の取得 */
Dcs *_get_dcs(long fd);
long __check_stat(const char *path);
int _io_observer(int sts, void *user_info);
int _io_writer(const char *buf, size_t size, void *user_info);
int _io_fd_buf_observer(int sts, void *user_info);
int _io_fd_buf_writer(const char *buf, size_t size, void *user_info);

/* 判定結果の取得 */
long dcs_io_is_decided() {
	return _is_decided;
}

/* 文字コード取得 */
long dcs_io_get_ctype() {
	return _stype;
}

/* 処理タイプ取得 */
long dcs_io_get_ope_type() {
	return _ope_type;
}

/* Dcs の取得 */
Dcs *_get_dcs(long fd) {
	if (fd < 0 || fd >= DCS_MAX_TIO) {
		return NULL;
	} else if (_tioInfo[fd].buf == NULL) {
		return NULL;
	}
	return _tioInfo[fd].dcs;
}

/* 文字コード判定（ファイル・パス） */
long dcs_io_path(const char *spath, long stype) {
	long sts;
	long sttype;
	Dcs *dcs;

	sttype = __check_stat(spath);
	if (sttype != _DST_FILE) {
	/* ファイルではない時 */
		return (sttype == _DST_NOTHING? DCS_ENOTEXISTS: DCS_ENOTFILE);
	}
	dcs = create_dcs();
	if (dcs == NULL) {
		return DCS_EALLOC;
	}
	sts = dcs_judges_file(dcs, NULL, NULL, spath, stype);
	release_dcs(dcs);
	return sts;
}

//------------------------------------------------------------

/* 処理ステータスからメッセージを取得 */
long dcs_io_statusto_message(char *str, long sts, long ctype) {
	int tsts;
	Dcs *dcs = create_dcs();

	if (dcs == NULL) {
		return DCS_EALLOC;
	}
	tsts = dcs_statusto_message(dcs, str, sts, ctype);
	release_dcs(dcs);
	return tsts;
}
/* コード・タイプから名前を取得 */
long dcs_io_typeto_name(char *name, long ctype) {
	return dcs_typeto_name(name, ctype);
}

/* コード名からタイプを取得 */
long dcs_io_nameto_type(const char *name) {
	return dcs_nameto_type(name);
}

/* 不特定のコード名を取得 */
long dcs_io_multi_type_names(char *names, long size, long ctype) {
	return dcs_multi_type_names(names, size, ctype);
}

/* すべてのコード名を取得 */
long dcs_io_all_type_names(char *names, long size) {
	return dcs_all_type_names(names, size);
}
/* すべてのコード名の長さを取得 */
long dcs_io_all_type_names_length() {
	return dcs_all_type_names_length();
}
//------------------------------------------------------------
/* エラー番号取得（dcs_io 共通）*/
long dcs_io_get_errno(long fd) {
	Dcs *dcs;

	_check_fd(fd);
	dcs = _get_dcs(fd);
	return dcs_get_errno(dcs);
}

/* エラー・メッセージ取得 */
long dcs_io_get_strerror(long fd, long err_no, char *str, long size, long ctype) {
	Dcs *dcs;

	_check_fd(fd);
	dcs = _get_dcs(fd);
	return dcs_get_strerror(dcs, err_no, str, size, ctype);
}

long _getTIO();

/* 入力モード判定 */
long dcs_io_is_can_read(long mode) {
	return ((mode == DMD_READ || mode == DMD_MODIFY)? 1: 0);
}

/* 出力モード判定 */
long dcs_io_is_can_write(long mode) {
	return ((mode == DMD_WRITE || mode == DMD_APPEND || mode == DMD_MODIFY)? 1: 0);
}

long _fd;

/* テキスト入出力オープン */
long dcs_io_open(const char *path, long mode, long stype, long dtype, bool is_forced, long *pstype) {
	long cmp, fd, sts;
	Dcs *dcs;

	if (pstype != NULL) {
		*pstype = 0;
	}
	/* アクセス・モードをチェック */
	switch(mode) {
	case DMD_READ:		/* 入力モード("rt" "rb") */
	case DMD_WRITE:		/* 出力モード("wt" "wb") */
	case DMD_APPEND:	/* 追加モード("at" "ab") */
	case DMD_MODIFY:	/* 更新モード("r+t" "r+b") */
		break;
	default:
		return DCS_EIOMODE;
	}
	_fd = fd = _getTIO();
	if (fd < 0) {
		return DCS_EIOMAX;
	}
	dcs = create_dcs();
	if (dcs == NULL) {
		dcs_io_close(fd);
		return DCS_EALLOC;
	}
	_tioInfo[fd].buf = malloc(_SBUF_SIZE);
	if (_tioInfo[fd].buf == NULL) {
		release_dcs(dcs);
		dcs_io_close(fd);
		return DCS_EALLOC;
	}
	_tioInfo[fd]._blength = _SBUF_SIZE;
	_tioInfo[fd].bsize = 0;
	_tioInfo[fd].dcs = dcs;
	_tioInfo[fd].mode = mode;
	_tioInfo[fd].shtype = STYPE_ASCII;
	_tioInfo[fd].stype = stype;
	_is_decided = false;
	_stype = DCS_ASCII;
	_ope_type = DOP_NOTHING;
	sts = dcs_convert_file(dcs, path, stype, dtype, is_forced, _io_fd_buf_observer, _io_fd_buf_writer, NULL);
	if (sts < 0) {
		dcs_io_close(fd);
		return sts;
	} else {
		if (DCS_CONFIRMED_CODE_TYPE(_stype)) {
		/* 確定できた時 */
			_is_decided = true;
		}
	}
	if (pstype != NULL) {
		*pstype = _stype;
	}

	if (_stype == dtype) {
	/* 入出力のコードが等しい時 */
		_ope_type = DOP_COPY;
	} else {
		if ((cmp = dcs_io_cmp_bom(_stype, dtype)) == 0) {
		/* コード変換の時 */
			_ope_type = DOP_CONV;
		} else {
			if (cmp > 0) {
			/* 同一 UTF コードで BOM 追加の時 */
				_ope_type = DOP_ADDBOM;
			} else {
			/* 同一 UTF コードで BOM 削除の時 */
				_ope_type = DOP_DELBOM;
			}
		}
	}
	_tioInfo[fd].dtype = dtype;
	dcs_io_rewind(fd);
	return fd;
} /* dcs_io_open */

/* テキスト入出力クローズ */
long dcs_io_close(long fd) {
	_check_fd(fd);
	if (_tioInfo[fd].dcs != NULL) {
		release_dcs(_tioInfo[fd].dcs);
		_tioInfo[fd].dcs = NULL;
	}
	free(_tioInfo[fd].buf);
	_tioInfo[fd].buf = NULL;
	return 0;
} /* dcs_io_close */

long _getTIO() {
	long i;

	for(i=0;i<DCS_MAX_TIO;i++) {
		if (_tioInfo[i].buf == NULL) {
			return i;
		}
	}
	return -1;
}

/* 入力ファイル巻き戻し */
long dcs_io_rewind(long fd) {
	_check_fd(fd);

	if (!dcs_io_is_can_read(_tioInfo[fd].mode)) {
		return DCS_EIOMODE;
	}
	_tioInfo[fd].bp = _tioInfo[fd].buf;
	return 0;
} /* dcs_io_rewind */

/* 変換後サイズ取得 */
long dcs_io_get_size(long fd) {
	_check_fd(fd);

	return _tioInfo[fd].bsize;
}

/* 入力文字コード取得 */
long dcs_io_get_stype(long fd) {
	_check_fd(fd);

	return _tioInfo[fd].stype;
}

/* 出力文字コード取得 */
long dcs_io_get_dtype(long fd) {
	_check_fd(fd);

	return _tioInfo[fd].dtype;
}

/* テキスト入出力１行入力（改行文字を除く）*/
long dcs_io_get_line(long fd, char *str, long size, long *psize) {
	long ccnt;
	size_t ssize, dsize;
	Dcs *dcs;

	_check_fd(fd);
	dcs = _get_dcs(fd);
	if (!dcs_io_is_can_read(_tioInfo[fd].mode)) {
		return DCS_EIOMODE;
	}
	if (_tioInfo[fd].bp >= _tioInfo[fd].buf + _tioInfo[fd].bsize) {
		return DCS_EOF;
	}
	ssize = (_tioInfo[fd].buf + _tioInfo[fd].bsize - _tioInfo[fd].bp);
	switch(_tioInfo[fd].dtype) {
	case DCS_ASCII:
	case DCS_JIS:
		ccnt = dcs_jis_strln(dcs, str, size, &dsize, &_tioInfo[fd].bp, ssize, &_tioInfo[fd].shtype);
		break;
	case DCS_SJIS:
		ccnt = dcs_sjis_strln(dcs, str, size, &dsize, &_tioInfo[fd].bp, ssize);
		break;
	case DCS_EUC:
	case DCS_EUC1:
	case DCS_EUC2:
	case DCS_EUC3:
		ccnt = dcs_euc_strln(dcs, str, size, &dsize, &_tioInfo[fd].bp, ssize);
		break;
	case DCS_UTF16BE:
	case DCS_UTF16BE_BOM:
		ccnt = dcs_unibe_strln(dcs, str, size, &dsize, &_tioInfo[fd].bp, ssize);
		break;
	case DCS_UTF16LE:
	case DCS_UTF16LE_BOM:
		ccnt = dcs_unile_strln(dcs, str, size, &dsize, &_tioInfo[fd].bp, ssize);
		break;
	case DCS_UTF8:
	case DCS_UTF8_BOM:
		ccnt = dcs_utf8_strln(dcs, str, size, &dsize, &_tioInfo[fd].bp, ssize);
		break;
	default:
		ccnt = DCS_ECTYPE;
		break;
	}
	if (psize != NULL) { // VBA 無効
		*psize = dsize;
	}
	return ccnt;
} /* dcs_io_get_line */

/*----------------------------------------------------------------------------*/

char *_dpath;
FILE *_dfp;

int _io_observer(int sts, void *user_info) {
	_stype = sts;
	switch(sts) {
	case DCS_ASCII:
	case DCS_JIS:
	case DCS_SJIS:
	case DCS_EUC:
	case DCS_EUC1:
	case DCS_EUC2:
	case DCS_EUC3:
	case DCS_UTF16BE:
	case DCS_UTF16BE_BOM:
	case DCS_UTF16LE:
	case DCS_UTF16LE_BOM:
	case DCS_UTF8:
	case DCS_UTF8_BOM:
	/* 確定できた時 */
		if ((_dfp = fopen(_dpath, "wb")) == NULL) {
			_stype = DCS_EWOPEN;
			return DCS_OBS_STOP;
		}
		_is_decided = true;
		break;
	default:
		if (!DCS_CONFIRMED_CODE_TYPE(sts)) {
		/* 入力タイプが確定できなかった時 */
			return DCS_OBS_STOP;
		}
		break;
	}
	return DCS_OBS_CONTINUE;
}

int _io_writer(const char *buf, size_t size, void *user_info) {
	if (fwrite(buf, 1, size, _dfp) != size) {
		return DCS_EWRITE;
	}
	return DCS_SUCCESS;
}

int _io_fd_buf_observer(int sts, void *user_info) {
	_tioInfo[_fd].stype = _stype = sts;
	switch(sts) {
	case DCS_ASCII:
	case DCS_JIS:
	case DCS_SJIS:
	case DCS_EUC:
	case DCS_EUC1:
	case DCS_EUC2:
	case DCS_EUC3:
	case DCS_UTF16BE:
	case DCS_UTF16BE_BOM:
	case DCS_UTF16LE:
	case DCS_UTF16LE_BOM:
	case DCS_UTF8:
	case DCS_UTF8_BOM:
	/* 確定できた時 */
		_is_decided = true;
		break;
	default:
		if (!DCS_CONFIRMED_CODE_TYPE(sts)) {
		/* 入力タイプが確定できなかった時 */
			return DCS_OBS_STOP;
		}
		break;
	}
	return DCS_OBS_CONTINUE;
}

int _io_fd_buf_writer(const char *buf, size_t size, void *user_info) {
	char *tbuffer;

	if (_tioInfo[_fd].bsize + size > _tioInfo[_fd]._blength) {
		tbuffer = malloc(_tioInfo[_fd]._blength * 2);
		if (tbuffer == NULL) {
			return DCS_EALLOC;
		}
		memcpy(tbuffer, _tioInfo[_fd].buf, _tioInfo[_fd].bsize);
		free(_tioInfo[_fd].buf);
		_tioInfo[_fd].buf = tbuffer;
		_tioInfo[_fd]._blength *= 2;
	}
	memcpy(_tioInfo[_fd].buf + _tioInfo[_fd].bsize, buf, size);
	_tioInfo[_fd].bsize += size;
	return DCS_SUCCESS;
}

long dcs_io_convert_file(const char *spath, long stype, const char *dpath, long dtype, bool is_forced) {
	long sts = DCS_SUCCESS;
	long cmp, rstype, sttype;
	Dcs *dcs;

	sttype = __check_stat(spath);
	if (sttype != _DST_FILE) {
	/* ファイルではない時 */
		return (sttype == _DST_NOTHING? DCS_ENOTEXISTS: DCS_ENOTFILE);
	}
	dcs = create_dcs();
	if (dcs == NULL) {
		return DCS_EALLOC;
	}
	_is_decided = false;
	_stype = DCS_ASCII;
	_ope_type = DOP_NOTHING;
	_dpath = (char *)dpath;
	_dfp = NULL;
	sts = dcs_convert_file(dcs, spath, stype, dtype, is_forced, _io_observer, _io_writer, NULL);
	if (_dfp != NULL) {
		fclose(_dfp);
	}
	if (sts != DCS_SUCCESS) {
		_stype = sts;
	} else {
		rstype = _stype;
		if (rstype == dtype) {
		/* 入出力のコードが等しい時 */
			_ope_type = DOP_COPY;
		} else if ((cmp = dcs_io_cmp_bom(stype, dtype)) == 0) {
			/* コード変換の時 */
			_ope_type = DOP_CONV;
		} else {
			if (cmp > 0) {
			/* 同一 UTF コードで BOM 追加の時 */
				_ope_type = DOP_ADDBOM;
			} else {
				/* 同一 UTF コードで BOM 削除の時 */
				_ope_type = DOP_DELBOM;
			}
		}
	}
	release_dcs(dcs);
	return sts;
} /* dcs_io_convert_file */

/* BOM 追加／削除判定 */
long dcs_io_cmp_bom(long stype, long dtype) {
	return dcs_cmp_bom(stype, dtype);
} /* dcs_io_cmp_bom */

long __check_stat(const char *path) {
    struct stat stbuf;

	if (stat(path, &stbuf) != 0) {
		return _DST_NOTHING;
	}
	return(S_ISREG(stbuf.st_mode)? _DST_FILE: S_ISDIR(stbuf.st_mode)? _DST_DIR: _DST_OTHER);
}
